//
//	MiscClipTextField.h -- a TextField subclass for clipping long string values
//		Written by Balazs Pataki Copyright (c) 1995 by Balazs Pataki.
//				Version 1.0.  All rights reserved.

#import <appkit/appkit.h>
#import	<misckit/MiscString.h>
#import <objc/objc-runtime.h>

#import "MiscClipTextField.h"
#import "MiscClipTextFieldCell.h"


#define	CLASS_NAME		"MiscClipTextField"
#define	CLASS_VERSION	1
#define CELL_CLASS		[MiscClipTextFieldCell class]


#define DELIMITERS		[delimiters stringValue]


static id	CellClass;	// Class variable of MiscClipTextField



/*
						********************************
						*                              *
						*      MiscClipTextField       *
						*                              *
						********************************
*/
/*
	DESCRIPTION:	...

	AUTHOR:			...

	NOTES:			...
*/

@implementation MiscClipTextField

+ initialize
// Set class version
{
	if (self == objc_lookUpClass(CLASS_NAME))  {
		[self setVersion:CLASS_VERSION];
		CellClass = CELL_CLASS;
	}
	return self;
}

+ setCellClass:classId
{
	CellClass=classId;
	return self;
}


- initFrame:(const NXRect *)rect
// Designated initializer
{
	id	oldCell;
	
	[super initFrame:rect];
										/* Plug in our own cell				*/
	oldCell = [self setCell:[[CellClass alloc] initTextCell:""]];
	[oldCell free];
	
	fullStringValue = [[MiscString allocFromZone:[self zone]] init];	
	// [[self cell] setClipEnabled:YES];
	
	return self;
}

- free
// Free our souls
{
	[fullStringValue free];
	
	return [super free];
}

- setClipEnabled:(BOOL)flag
// Sets whether the next `setStringValue:' message should clip the text or not 
{
	[[self cell] setClipEnabled:flag];

	return self;
}



- setClipOnRight:(BOOL)flag
// If flag is YES clipping happens on the right, otherwise on the left of the 
// string in the cell. The string value is updated and redisplayed according to 
// the value of `flag'.
{
	[[self cell] setClipOnRight:flag];
	//[self setStringValue:[fullStringValue stringValue]];

	return self;
}


- setClipperString:(const char*)aString
// Sets  `aString' as the string that is displayed in place of the clipped part
// of the original string. The string value is updated and redisplayed 
// according to the value of `aString'.
{
	[[self cell] setClipperString:aString];
	//[self setStringValue:[fullStringValue stringValue]];

	return self;
}


- clipByChars:(const char*)delimChars
// Sets  `delimChars' as the delimiters by which the clipping has to happen. 
// If `delimChars' is NULL or an empty string clipping will occur to any 
// character in the string. ( Don, Could you please describe this better for 
// me! I just can't find the right words :-( 
// The string value is updated and redisplayed according to the value of 
// `delimChars'.
{
	[[self cell] setClipDelimiters:delimChars];
	//[self setStringValue:[fullStringValue stringValue]];

	return self;
}


- setStringValue:(const char *)aString
{
										/*	Save the new string value		*/
	[fullStringValue setStringValue:aString];	
										/* Super sends a message to our cell*/
										/*	to display the string value		*/
	return [super setStringValue:aString];

}

- takeStringValueFrom:sender
{
	[self setStringValue:[sender stringValue]];
	return self;
}

- sizeTo:(NXCoord)width :(NXCoord)height
// Adjusts the current text to the new size and redisplays it
{
	[super sizeTo:width :height];
	[self setStringValue:[fullStringValue stringValue]];
	return self;
}


- mouseDown:(NXEvent *)event
// In case of triple-click event displays the full text instead of the clipped
// one and allows selecting and scrolling the string
// NOTE: This trick alters the alignment of the cell and changes it to 
//		 align its text to left and to truncate the text - don't know any
//		 workaround (maybe we should stay with left aligned text only :-(
{ 
	// Handling only triple-click event
	if (event->data.mouse.click == 3) {
		id	theCell  = [self cell];		
		[[self window] disableDisplay];
		[theCell setSelectable:YES];
		[theCell setScrollable:YES];
		[self setClipEnabled:NO]; 		/* Turn off clipping for a moment	*/
		[self setStringValue:[fullStringValue stringValue]];
		[self setClipEnabled:YES];		/* ... and turn it back				*/
		[[self window] reenableDisplay];
		[self display];
	}
	return [super mouseDown:event];
}

- (const char*)fullStringValue;
{
	return [fullStringValue stringValue];
}


/*
- setAlignment:(int)mode
{
	[[self window] disableFlushWindow];
	
	[super setAlignment:mode];
	[self setStringValue:[fullStringValue stringValue]];
	
	[[[self window] reenableFlushWindow] flushWindow];

	return self;
}
*/

- clipper					{ return [[self cell] clipper];			}
- delimiters				{ return [[self cell] delimiters];		}
- (BOOL) doesClipOnRight	{ return [[self cell] doesClipOnRight];	}
- (BOOL) isClipEnabled		{ return [[self cell] isClipEnabled];	}


- read:(NXTypedStream *)stream
{
	[super read:stream];
	
	fullStringValue = NXReadObject(stream);
	
	return self;
}


- write:(NXTypedStream *)stream
{
	[super write:stream];
	
	NXWriteObject(stream, fullStringValue);
	
	return self;
}


- awake 
{
	return self;
}

@end


/*
							********************************
							*                              *
							*  MiscClipTextField(IBStuff)  *
							*                              *
							* -- InterfaceBuilder Stuff--  *
							*                              *
							********************************
*/

@implementation MiscClipTextField(IBStuff)

- (const char *)getInspectorClassName
// Return the class name of our inspector.
{
	return "MiscClipTextFieldInspector";
}

@end


/*
							********************************
							*                              *
							*   MiscClipTextField(Test)    *
							*                              *
							* -- Debug and Test Methods -- *
							*                              *
							********************************
*/

@implementation MiscClipTextField(Test)

- changeFont:sender
{
	[[self cell] setFont:[sender selFont]];
	[self setStringValue:[fullStringValue stringValue]];
	return self;
}

@end

