//
//	PBPClipTextFieldCell.m -- used for implementing PBPClipTextField
//		Written by Balazs Pataki Copyright (c) 1995 by Balazs Pataki.
//				Version 1.0.  All rights reserved.

#import <appkit/appkit.h>
#import	<misckit/MiscString.h>
#import <objc/objc-runtime.h>
#import "PBClipTextFieldCell.h"


#define	CLASS_NAME		"PBClipTextFieldCell"
#define	CLASS_VERSION	1


#define	DEFAULTCLIPPER	"..."
#define DELIMITERS		[delimiters stringValue]


// The width of the text displayed in a Cell seems shorter than the view the 
// Cell is displayed in so a bit of adjustment is reasonable
#define TRUNCATION_ADJUSTMENT	3.0

#define FLUSH 	NXPing();											\
			[[[[self controlView] window] reenableFlushWindow] flushWindow];

/*
#define PRINT 	puts([self stringValue]);									\
				printf("Control's width:  %f   height:%f\n", 				\
					NX_WIDTH(&viewRect), NX_HEIGHT(&viewRect));				\
				printf("Calc'd cell width:%f   height:%f\n", 				\
					size.width, size.height);								\
				printf("Text width:  %f Last char width: %f\n",				\ 
					[theFont getWidthOf:[temp stringValue]], lastCharSize);	\
*/

@interface PBClipTextFieldCell (Private)
- _clipStringValue:(const char *)aString;
@end


/*
						********************************
						*                              *
						*     PBClipTextFieldCell      *
						*                              *
						********************************
*/
/*
	DESCRIPTION:	...

	AUTHOR:			...

	NOTES:			...
*/

@implementation PBClipTextFieldCell

+ initialize
// Set class version
{
	if (self == objc_lookUpClass(CLASS_NAME))  {
		[self setVersion:CLASS_VERSION];
	}
	return self;
}

- init;
// Initializes a newly allocated PBClipTextFieldCell with default values, that
// is the clipper string to "...", uses no delimiters when clipping, and
// clipping happens on the right. The cell's other attributes are set for
// displaying text only (no scroll, selection allowed).  
{
	[super init];

	clipper = [[MiscString allocFromZone:[self zone]] 
						   initString:DEFAULTCLIPPER];
	delimiters  = nil;				/* As a deafult we don't have delimiters*/
	clipOnRight = YES;				/*	and clipping happens on the right	*/
	clipEnabled = YES;
	
	// MiscClipField is initialized for only displaying its string value 
	// - is it OK?
	[self setBezeled:YES];
	[self setEditable:NO];
	[self setScrollable:NO];
	[self setBackgroundGray:NX_LTGRAY];

	return self;
}

- free
{
	[clipper free];
	if (delimiters)
		[delimiters free];
	
	return [super free];
}

- setClipOnRight:(BOOL) flag
// If flag is YES clipping happens on the right, otherwise on the left of the 
// string in the cell
{
	clipOnRight = flag;

	return self;
}


- setClipperString:(const char*) aString
// Sets  `aString' as the string that is displayed in place of the clipped part
// of the original string
{
	[clipper setStringValue:aString];

	return self;
}


- setClipDelimiters:(const char*) delimChars
// Sets  `delimChars' as delimiters by which the clipping has to happen
{
	if (delimiters) 
		[delimiters setStringValue:delimChars];
	else
		delimiters = [MiscString newWithString:delimChars];

	return self;
}

- setClipEnabled:(BOOL) flag
// Sets whether the next `setStringValue:' message should clip the text or not 
{
	clipEnabled = flag;
	return self;
}


- setStringValue:(const char *)aString
{
	if ( clipEnabled )
		return [self _clipStringValue:aString];
	
	return [super setStringValue:aString];
	
}

- clipper		{ return clipper;	 }
- delimiters	{ return delimiters; }	

- (BOOL) isWrapped
// Returns yes if the Cell wraps the text by word
{
	return ( cFlags2.noWrap ? NO : YES );
}

- (BOOL) isClipEnabled		{ return clipEnabled; }
- (BOOL) doesClipOnRight	{ return clipOnRight; }

- write:(NXTypedStream *)stream
{
    [super write:stream];
    
    NXWriteObject(stream, clipper);
    NXWriteObject(stream, delimiters);
    NXWriteType(stream, @encode(BOOL), &clipOnRight);
    NXWriteType(stream, @encode(BOOL), &clipEnabled);
   
    return self;
}


- read:(NXTypedStream *)stream
{
    [super read:stream];
   
	clipper = NXReadObject(stream);
	delimiters = NXReadObject(stream);
	NXReadType(stream, @encode(BOOL), &clipOnRight);
	NXReadType(stream, @encode(BOOL), &clipEnabled);

	return self;
}


- awake
{
	return self;
}




@end


/*
						********************************
						*                              *
						*   PBClipTextField(Private)   *
						*                              *
						*    -- Private Methods --     *
						*                              *
						********************************
*/
/*
	NOTES:	`_clipStringValue:' could be implemented as a function which would
			make it possible to use this clipping feature in other custom
			cell classes (e.g. MyClippingBrowserCell) as well. What do you 
			think ? 
			
	char* clipString(string, delimiters, clipper, font, fromRight, maxWidth)
		  			const char 	*string;	//The string to clip
					const char  *delimiters;//Delimiters to clip by
					const char 	*clipper;	//String that subst. invis. part 
					id			font;		//Font the string is displayed with
					BOOL		fromRight;	//Clipping from left or right ?	
					float 		maxWidth; 	//The max width of the object the
											//string will be displayed in e.g.
											//an adjusted Cell width

			Or maybe better to include it in a category of ActionCell thus 
			applying clipping option to all ActionCell subclasses (it has to be 
			an ActionCell category beacause it uses `controlView' to 
			determine the size of the view the cell is displayed in). But we 
			need the clipper and delimiter arguments then, don't we.
			So, it could look like:


	@interface ActionCell(ClippingText)
	
	- clipStringValue:(const char *) theString
		fromRight:(BOOL)yesno
		usingDelimiters:(const char*) delimiters
		clipperString:(const char*) clipperString
	
	@end;
	
			Which one do you think is better?
	
*/


@implementation PBClipTextFieldCell (Private)

- _clipStringValue:(const char *)aString
// A very long story...
// 
{
	id	temp 	 = 	nil;
	id 	mainFont = 	[self font];
	id 	scrnFont = 	[mainFont screenFont];							
	id 	theFont  = 	(scrnFont ? scrnFont : mainFont); // try to use screen font
	float 	clipperWidth =	0; 
	float 	clippedWidth =	0;
	float 	cellWidth	 =  0;
	BOOL  	usingDelimiters;
	NXRect	viewRect;
	NXSize	size;
	
	if (!aString) 
		return self;
											/* More initializing			*/		
	clippedWidth = [theFont getWidthOf:aString];
	temp = [[MiscString alloc] initString:aString];

	if (![clipper stringValue])
		[clipper setStringValue:DEFAULTCLIPPER];

	clipperWidth =	[theFont getWidthOf:[clipper stringValue]]; 

	[[[self controlView] window] disableFlushWindow];
	
	
	// Figure out the initial cell width (the biggest possible cell width)
	[[self controlView] getFrame:&viewRect];
	cellWidth =	NX_WIDTH(&viewRect);
										/* Initial cell "width" depends on 	*/
										/*	our border as well...			*/
	if ([self isBezeled])
		cellWidth -= 3.0;
	else if ([self isBordered])
		cellWidth -= 2.0;
	
							// Wee need this little adjustment - don't really
							// know why, but we do
	if ( ![self isWrapped] )
		cellWidth -= TRUNCATION_ADJUSTMENT;
	
	usingDelimiters = ( delimiters && ![delimiters emptyString] ? YES : NO );

								
	[self setSelectable:NO];			/* Disable selecting and scrolling	*/
	[self setScrollable:NO];
	
	// Clip on the right if `aString' is too long
	if ((clippedWidth > cellWidth) && clipOnRight) {
		BOOL cutMore=NO;	// This becomes true if text doesn't fits into 
							// the cell in the first round 
						/* The loop cuts characters from the end and checks	*/
						/*	wheter it's short enough to be displayed		*/
		while (YES) {
			clippedWidth=[theFont getWidthOf:[temp stringValue]];
			
								/* If the string is still too long			*/
			if ( ((clippedWidth+clipperWidth) >= cellWidth) || cutMore)  {
				[temp removeFrom:[temp length]-1 length:1];
				if ( usingDelimiters ) {
								/* No delimiter found means we reached the 	*/
								/*	last delimited string element and it 	*/
								/*  can't be shortened anymore because there*/
								/*	is no "next" delimter backwards			*/
					int from = [temp rspotOfChars:DELIMITERS];
					if ( from == -1 )
						[temp replace:[temp stringValue] with:""];
					else
						[temp removeFrom:from+1 length:[temp length]-1];
				}
								/* Last element reached and still long		*/
				if ([temp emptyString]) {
					[temp setStringValue:[clipper stringValue]];
					FLUSH;
					return [super setStringValue:[temp stringValueAndFree]];
				}
				cutMore = NO;
			} //Enf if (still too long)
			else {
				[temp concatenate:clipper];
									/* Set the clipped string value 		*/
				[super setStringValue:[temp stringValue]];	
									/* Calc how much space it needed to be	*/
									/*  displayed							*/
				[self calcCellSize:&size inRect:&viewRect];
									/* It didn't fit so its broken into two */
									/* lines (ie: the cell got higher) 		*/
				if ([self isWrapped] && (size.height > NX_HEIGHT(&viewRect))){
					cutMore = YES;	/* Force cutting some more				*/
									/* Remove clipper string from end		*/
					[temp removeFrom:([temp length]-[clipper length]) 
						  length:[clipper length]];
				}			
				else {			/* Finished, string value can be displyed*/
					[temp free];
					FLUSH;
					return self;
				}
			} // End else (check if string value can really be displayed)
		} // End of string shortening loop
	}//End if (clipOnRight)



	// Clip on the left if `aString' is too long
	// (This code is nearly identical to the right clipping one, except that
	//  it scans the string from left to right. Can you see any way to merge 
	//  these two in a meaningful way?) 
	if ( (clippedWidth > cellWidth) && !clipOnRight) {
		BOOL cutMore=NO;	// This becomes true if text doesn't fits into 
							// the cell in the first round 		
						/* The loop cuts characters from the beg. of temp	*/
						/*	and checks	wheter it's short enough to be 		*/
						/*	displayed										*/
		while ( YES ) {
						// This sometimes lies, maybe because of the screenfont
						// - printerfont differency
			clippedWidth=[theFont getWidthOf:[temp stringValue]];
								/* If string is still too long	or we are	*/
								/*	forced to cut more ...					*/
			if ( ((clippedWidth+clipperWidth) >= cellWidth) || cutMore)  {
				[temp removeFrom:0 length:1];
				if ( usingDelimiters ) {
					int to=[temp spotOfChars:DELIMITERS];
								/* No delimiters found means we reached the */
								/*	last delimited string element and it 	*/
								/*  can't be shortened anymore because there*/
								/*	is no "next" delimter forward			*/
					if (to == -1)
						[temp replace:[temp stringValue] with:""];
					else
						[temp removeFrom:0 to:to-1];
				}
								/* Last element reached and still long		*/
				if ([temp emptyString]) {
					[temp setStringValue:[clipper stringValue]];
					FLUSH;
					return [super setStringValue:[temp stringValueAndFree]];
				}
				cutMore = NO;	/* Turn forcing cutting more off			*/
			} //Enf if (still too long)
			else {	
									/* Insert clipper						*/
				[temp insertString:clipper];
									/* Set the clipped string value 		*/
				[super setStringValue:[temp stringValue]];	
									/* Calc how much space it needed to be	*/
									/*  displayed							*/
				[self calcCellSize:&size inRect:&viewRect];
				
				if ([self isWrapped] && (size.height > NX_HEIGHT(&viewRect))){
									/* It didn't fit so its broken into many*/
									/* lines (ie: the cell got higher) 		*/
					cutMore = YES;	/* Force cutting some more				*/
									/* Remove clipper string from front		*/
					[temp removeFrom:0 length:[clipper length]];
				}			
				else {			/* Finished, string value can be displyed*/
					[temp free];
					FLUSH;
					return self;
				}
			} // End else (check if string value can really be displayed)
		} // End of string shortening loop
	}//End if (clip on left)

	FLUSH;
	return [super setStringValue:[temp stringValueAndFree]];
	
}



@end

